using System;
using System.IO;
using System.Collections.Generic;
using System.Globalization;
using CG2.Mathematics;

namespace CG2.Modeling
{
    public class ObjMeshLoader
    {
        public static NumberStyles styles = NumberStyles.Integer | NumberStyles.AllowDecimalPoint;
        public static CultureInfo provider = new CultureInfo("en-US");

        //Summary
        public static bool Load(ObjMesh mesh, string fileName)
        {
            try
            {
                using (StreamReader streamReader = new StreamReader(fileName))
                {
                    Load(mesh, streamReader);
                    streamReader.Close();
                    return true;
                }
            }
            catch { return false; }
        }

        static char[] splitCharacters = new char[] { ' ' };

        static List<Vector4> vertices;
        static List<Vector4> normals;
        static List<Vector4> texCoords;
        static Dictionary<ObjMesh.ObjVertex, int> objVerticesIndexDictionary;
        static List<ObjMesh.ObjVertex> objVertices;
        static List<ObjMesh.ObjTriangle> objTriangles;

        static void Load(ObjMesh mesh, TextReader textReader)
        {
            vertices = new List<Vector4>();
            normals = new List<Vector4>();
            texCoords = new List<Vector4>();
            objVerticesIndexDictionary = new Dictionary<ObjMesh.ObjVertex, int>();
            objVertices = new List<ObjMesh.ObjVertex>();
            objTriangles = new List<ObjMesh.ObjTriangle>();

            string line;
            while ((line = textReader.ReadLine()) != null)
            {
                line = line.Trim(splitCharacters);
                line = line.Replace("  ", " ");

                string[] parameters = line.Split(splitCharacters);

                switch (parameters[0])
                {
                    case "p": // Point
                        break;

                    case "v": // Vertex
                        float x;
                        float.TryParse(parameters[1], styles, provider, out x);
                        float y;
                        float.TryParse(parameters[2], styles, provider, out y);
                        float z;
                        float.TryParse(parameters[3], styles, provider, out z);
                        vertices.Add(new Vector4(x, y, z));
                        break;

                    case "vt": // TexCoord
                        float u;
                        float.TryParse(parameters[1], styles, provider, out u);
                        float v;
                        float.TryParse(parameters[2], styles, provider, out v);
                        texCoords.Add(new Vector4(u, v, 0));
                        break;

                    case "vn": // Normal
                        float nx;
                        float.TryParse(parameters[1], styles, provider, out nx);
                        float ny;
                        float.TryParse(parameters[2], styles, provider, out ny);
                        float nz;
                        float.TryParse(parameters[3], styles, provider, out nz);
                        normals.Add(new Vector4(nx, ny, nz));
                        break;

                    case "f":
                        switch (parameters.Length)
                        {
                            case 4:
                                ObjMesh.ObjTriangle objTriangle = new ObjMesh.ObjTriangle();
                                objTriangle.Index0 = ParseFaceParameter(parameters[1]);
                                objTriangle.Index1 = ParseFaceParameter(parameters[2]);
                                objTriangle.Index2 = ParseFaceParameter(parameters[3]);
                                objTriangles.Add(objTriangle);
                                break;
                        }
                        break;
                }
            }

            mesh.Vertices = objVertices.ToArray();
            mesh.Triangles = objTriangles.ToArray();

            objVerticesIndexDictionary = null;
            vertices = null;
            normals = null;
            texCoords = null;
            objVertices = null;
            objTriangles = null;
        }

        static char[] faceParamaterSplitter = new char[] { '/' };
        static int ParseFaceParameter(string faceParameter)
        {
            Vector4 vertex = new Vector4();
            Vector4 texCoord = new Vector4();
            Vector4 normal = new Vector4();

            string[] parameters = faceParameter.Split(faceParamaterSplitter);

            int vertexIndex;
            int.TryParse(parameters[0], styles, provider, out vertexIndex);
            if (vertexIndex < 0) vertexIndex = vertices.Count + vertexIndex;
            else vertexIndex = vertexIndex - 1;
            vertex = vertices[vertexIndex];

            if (parameters.Length > 1)
            {
                int texCoordIndex;
                int.TryParse(parameters[1], styles, provider, out texCoordIndex);
                if (texCoordIndex < 0) texCoordIndex = texCoords.Count + texCoordIndex;
                else texCoordIndex = texCoordIndex - 1;
                texCoord = texCoords[texCoordIndex];
            }

            if (parameters.Length > 2)
            {
                int normalIndex;
                int.TryParse(parameters[2], styles, provider, out normalIndex);
                if (normalIndex < 0) normalIndex = normals.Count + normalIndex;
                else normalIndex = normalIndex - 1;
                normal = normals[normalIndex];
            }

            return FindOrAddObjVertex(ref vertex, ref texCoord, ref normal);
        }

        static int FindOrAddObjVertex(ref Vector4 vertex, ref Vector4 texCoord, ref Vector4 normal)
        {
            ObjMesh.ObjVertex newObjVertex = new ObjMesh.ObjVertex();
            newObjVertex.Vertex = vertex;
            newObjVertex.TexCoord = texCoord;
            newObjVertex.Normal = normal;

            int index;
            if (objVerticesIndexDictionary.TryGetValue(newObjVertex, out index))
            {
                return index;
            }
            else
            {
                objVertices.Add(newObjVertex);
                objVerticesIndexDictionary[newObjVertex] = objVertices.Count - 1;
                return objVertices.Count - 1;
            }
        }
    }
}