﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using System.Drawing;
using CG2.Modeling;
using CG2.Shading;
using CG2.Lighting;

namespace CG2.Rendering
{
	public class Camera
	{
		#region Properties

		public Vector4 Position;
		public Vector4 Target;
		public Vector4 Up = new Vector4(0, 0, 1);
		public Double FovY = 45;
		public Vector4 U, V, W; 
		public Vector4 BgColor = new Vector4(0, 0, 0);

        public RayAccelerator Accelerator = new RayAccelerator();
		public World World;

		public Int32 Width;
		public Int32 Height;
		public Vector4[] Pixels;

		public Bitmap Bitmap;
        // Native size of Bitmap (BitmapWidth, BitmapHeight)
        public int BitmapWidth, BitmapHeight; 

        // Maximal intensity of calculated color (from shader)
        public Double MaxIntensity = 10;

        // SubSampleCount (=1) being the number of sub sampling rays per pixel
        public int SubSampleCount = 1;

        // TODO: declare UseShadows to control shadow rendering
        public Boolean UseShadows = false;

		#endregion

		#region Init

        public Camera(Int32 bitmapWidth, Int32 bitmapHeight)
		{
            BitmapWidth = bitmapWidth;
            BitmapHeight = bitmapHeight;
		}

		#endregion

		#region Buffer Acess

		public Vector4 GetPixel(Int32 i, Int32 j)
		{
			return Pixels[i + j * Width];
		}

		public void SetPixel(Int32 i, Int32 j, Vector4 color)
		{
			Pixels[i + j * Width] = color;
		}

		#endregion

		#region Rendering

        /// <summary>Render has three parts:
        /// <para>1. Build the structure - use data from geometry</para> 
        /// <para>2. Trace color values for rays to the pixel buffer. Place for Lights, Textures etc. to take an effect.</para>
        /// <para>3. Copy results from the pixel buffer to the bitmap image. Place to post-processing effects - AA, DOF, Blur</para>
        /// </summary>
		public void Render()
		{
            Accelerator.BuildStructure(World);

            InitBuffers();

			RayTrace();
			
            PresentFrame();
		}

        public void InitBuffers()
        {
            // Set total number of pixels (Width, Height) in Pixels buffer based on the SubSampleCount and size of bitmap
            Width = BitmapWidth * SubSampleCount;
            Height = BitmapHeight * SubSampleCount;
            // Create bitmap 
            Int32 pixelCount = Width * Height;
            Bitmap = new Bitmap(BitmapWidth, BitmapHeight);
            Pixels = new Vector4[pixelCount];
        }

        /// <summary>Derived from Computer Graphics - David Mount.
        /// <para>Implementations can differ from another - make your own, from scratch. </para>
        /// <para>F.e., use blender to see the difference during the change of fovy value.</para>
        /// <para>See http://goo.gl/q6Sz0 and http://goo.gl/rB8J6 </para>
        /// </summary>
		public void RayTrace()
		{
            W = Position - Target;
			U = Up % W;
			V = W % U;

			U = U.Normalized;
			V = V.Normalized;
			W = W.Normalized;

			Double x0 = (1.0 - Width) / 2;
			Double y0 = (1.0 - Height) / 2;
			Double z = Height / (2 * Math.Tan(MathEx.DegToRad(FovY) / 2));

			Double y = y0;
			for (Int32 j = Height - 1; j >= 0; j--, y += 1.0f)
			{
				Double x = x0;
				for (Int32 i = 0; i < Width; i++, x += 1.0f)
				{
                    Vector4 rayDir = (x * U + y * V - z * W).Normalized;
					Ray ray = new Ray(Position, rayDir);
					Vector4 color = RayTrace(ray);

					SetPixel(i, j, color);
				}
			}
		}

		public Vector4 RayTrace(Ray ray)
		{
            Vector4 color = Vector4.Zero;

            Accelerator.Collide(ray);

			if (ray.HitModel == null) return BgColor; // no collision return

            Vector4 hitPoint = ray.GetHitPoint();
            Ray lightRay = new Ray();
            Vector4 eyeDir = -ray.Direction;
            foreach (Light light in World.Lights)
            {
                light.SetLightRayAt(hitPoint, lightRay);
                Double nl = ray.HitNormal * lightRay.Direction;
                if (UseShadows && nl > 0.0f) Accelerator.Collide(lightRay);

                Double intensityFactor = (lightRay.HitModel != null) ? 0 : 1;

                color += ray.HitModel.Shader.GetColor(hitPoint, ray.HitNormal, eyeDir, lightRay.Direction, intensityFactor * light.GetIntensityAt(hitPoint), light);
            }

            return color;
		}

		public void PresentFrame()
		{
            for (Int32 j = 0; j < BitmapHeight; j++)
            {
                for (Int32 i = 0; i < BitmapWidth; i++)
                {
                    Vector4 c = Vector4.Zero;
                    // calculated sub-sampled color by averaging colors of sub-pixels
                    for (Int32 jSub = 0; jSub < SubSampleCount; jSub++)
                    {
                        for (Int32 iSub = 0; iSub < SubSampleCount; iSub++)
                        {
                            c += GetPixel(i * SubSampleCount + iSub, j * SubSampleCount + jSub);
                        }
                    }
                    c *= (1.0 / (double)(SubSampleCount * SubSampleCount));

                    c = Vector4.Clamp(c, 0, 1);

                    Bitmap.SetPixel(i, j, Color.FromArgb(
                        (Int32)(255 * c.X),
                        (Int32)(255 * c.Y),
                        (Int32)(255 * c.Z))
                    );
                }
            }
		}

		#endregion
	}
}
