﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using CG2.Lighting;
using CG2.Texturing;

namespace CG2.Shading
{
	public class PhongShader : Shader
	{
		#region Properties

        // Replace DiffuseColor with DiffuseSampler
        // Replace SpecularColor with SpecularSampler
        // Replace SpecularColor with SpecularSampler
        public Sampler DiffuseSampler = new ColorSampler(new Vector4(1.0, 1.0, 1.0));
        public Sampler SpecularSampler = new ColorSampler(new Vector4(1.0, 1.0, 1.0));
        public Sampler AmbientSampler = new ColorSampler(new Vector4(0.1, 0.1, 0.1));

		public Double Shininess = 20;

		#endregion


		#region Init

        //Hint: We can create phong shader with different parameters
        //      Default is phong with just a color
        //      Try to change specular and ambient color to see the behavior
		public PhongShader()
		{
		}

		public PhongShader(Vector4 diffuseColor)
		{
            DiffuseSampler = new ColorSampler(diffuseColor);
		}

		public PhongShader(Vector4 diffuseColor, Vector4 specularColor)
		{
            DiffuseSampler = new ColorSampler(diffuseColor);
            SpecularSampler = new ColorSampler(specularColor);
		}

		public PhongShader(Vector4 diffuseColor, Vector4 specularColor, Vector4 ambientColor)
		{
            DiffuseSampler = new ColorSampler(diffuseColor);
            SpecularSampler = new ColorSampler(specularColor);
            AmbientSampler = new ColorSampler(ambientColor);
		}

		#endregion


		#region Shading

        /// <summary>
        /// Blinn-Phong local shading model
        /// </summary>
        /// <param name="point"></param>
        /// <param name="normal"></param>
        /// <param name="eyeDir"></param>
        /// <param name="lightDir"></param>
        /// <param name="lightColor"></param>
        /// <returns></returns>
        public override Vector4 GetColor(Vector4 point, Vector4 normal, Vector4 eyeDir, Vector4 lightDir, Double lightIntensity, Light light)
        {
            Double diffuseFactor = lightIntensity * (normal * lightDir);
            if (diffuseFactor < 0) diffuseFactor = 0;

            Vector4 half = (lightDir + eyeDir).Normalized;
            Double specularFactor = lightIntensity * Math.Pow(normal * half, Shininess);

            // Get sample of DiffuseSampler at texture coordinates calculated with DiffuseMapping at given point
            Vector4 diffuseColor = DiffuseSampler.GetSample(point);

            // Get sample of SpecularSampler at texture coordinates calculated with SpecularMapping at given point
            Vector4 specularColor = SpecularSampler.GetSample(point);

            // Get sample of AmbientSampler at texture coordinates calculated with AmbientMapping at given point
            Vector4 ambientColor = AmbientSampler.GetSample(point);

            Vector4 color = new Vector4();

            color = diffuseFactor * (diffuseColor ^ light.DiffuseColor);
            color += specularFactor * (specularColor ^ light.SpecularColor);
            color += ambientColor ^ light.AmbientColor;

            return color;
		}

		#endregion
	}
}
